/**
	\file
	\brief Contains OpenCL::Exception methods.
*/




#include "settings.h"




#if ( SETTING_OPENCL != 0 )




#include "opencl_exception.h"
#include "helpers.h"

#include <sstream>




namespace OpenCL {




/*==============================================================================
	Exception methods
==============================================================================*/


Exception::Exception( char const* const file, unsigned int const line, std::string const& what, cl_int const code ) {

	char const* message = "<unknown error>";
	switch( code ) {
		case CL_SUCCESS:                                   { message = "CL_SUCCESS";                                   break; }
		case CL_DEVICE_NOT_FOUND:                          { message = "CL_DEVICE_NOT_FOUND";                          break; }
		case CL_DEVICE_NOT_AVAILABLE:                      { message = "CL_DEVICE_NOT_AVAILABLE";                      break; }
		case CL_COMPILER_NOT_AVAILABLE:                    { message = "CL_COMPILER_NOT_AVAILABLE";                    break; }
		case CL_MEM_OBJECT_ALLOCATION_FAILURE:             { message = "CL_MEM_OBJECT_ALLOCATION_FAILURE";             break; }
		case CL_OUT_OF_RESOURCES:                          { message = "CL_OUT_OF_RESOURCES";                          break; }
		case CL_OUT_OF_HOST_MEMORY:                        { message = "CL_OUT_OF_HOST_MEMORY";                        break; }
		case CL_PROFILING_INFO_NOT_AVAILABLE:              { message = "CL_PROFILING_INFO_NOT_AVAILABLE";              break; }
		case CL_MEM_COPY_OVERLAP:                          { message = "CL_MEM_COPY_OVERLAP";                          break; }
		case CL_IMAGE_FORMAT_MISMATCH:                     { message = "CL_IMAGE_FORMAT_MISMATCH";                     break; }
		case CL_IMAGE_FORMAT_NOT_SUPPORTED:                { message = "CL_IMAGE_FORMAT_NOT_SUPPORTED";                break; }
		case CL_BUILD_PROGRAM_FAILURE:                     { message = "CL_BUILD_PROGRAM_FAILURE";                     break; }
		case CL_MAP_FAILURE:                               { message = "CL_MAP_FAILURE";                               break; }
		case CL_MISALIGNED_SUB_BUFFER_OFFSET:              { message = "CL_MISALIGNED_SUB_BUFFER_OFFSET";              break; }
		case CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST: { message = "CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST"; break; }
		case CL_INVALID_VALUE:                             { message = "CL_INVALID_VALUE";                             break; }
		case CL_INVALID_DEVICE_TYPE:                       { message = "CL_INVALID_DEVICE_TYPE";                       break; }
		case CL_INVALID_PLATFORM:                          { message = "CL_INVALID_PLATFORM";                          break; }
		case CL_INVALID_DEVICE:                            { message = "CL_INVALID_DEVICE";                            break; }
		case CL_INVALID_CONTEXT:                           { message = "CL_INVALID_CONTEXT";                           break; }
		case CL_INVALID_QUEUE_PROPERTIES:                  { message = "CL_INVALID_QUEUE_PROPERTIES";                  break; }
		case CL_INVALID_COMMAND_QUEUE:                     { message = "CL_INVALID_COMMAND_QUEUE";                     break; }
		case CL_INVALID_HOST_PTR:                          { message = "CL_INVALID_HOST_PTR";                          break; }
		case CL_INVALID_MEM_OBJECT:                        { message = "CL_INVALID_MEM_OBJECT";                        break; }
		case CL_INVALID_IMAGE_FORMAT_DESCRIPTOR:           { message = "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR";           break; }
		case CL_INVALID_IMAGE_SIZE:                        { message = "CL_INVALID_IMAGE_SIZE";                        break; }
		case CL_INVALID_SAMPLER:                           { message = "CL_INVALID_SAMPLER";                           break; }
		case CL_INVALID_BINARY:                            { message = "CL_INVALID_BINARY";                            break; }
		case CL_INVALID_BUILD_OPTIONS:                     { message = "CL_INVALID_BUILD_OPTIONS";                     break; }
		case CL_INVALID_PROGRAM:                           { message = "CL_INVALID_PROGRAM";                           break; }
		case CL_INVALID_PROGRAM_EXECUTABLE:                { message = "CL_INVALID_PROGRAM_EXECUTABLE";                break; }
		case CL_INVALID_KERNEL_NAME:                       { message = "CL_INVALID_KERNEL_NAME";                       break; }
		case CL_INVALID_KERNEL_DEFINITION:                 { message = "CL_INVALID_KERNEL_DEFINITION";                 break; }
		case CL_INVALID_KERNEL:                            { message = "CL_INVALID_KERNEL";                            break; }
		case CL_INVALID_ARG_INDEX:                         { message = "CL_INVALID_ARG_INDEX";                         break; }
		case CL_INVALID_ARG_VALUE:                         { message = "CL_INVALID_ARG_VALUE";                         break; }
		case CL_INVALID_ARG_SIZE:                          { message = "CL_INVALID_ARG_SIZE";                          break; }
		case CL_INVALID_KERNEL_ARGS:                       { message = "CL_INVALID_KERNEL_ARGS";                       break; }
		case CL_INVALID_WORK_DIMENSION:                    { message = "CL_INVALID_WORK_DIMENSION";                    break; }
		case CL_INVALID_WORK_GROUP_SIZE:                   { message = "CL_INVALID_WORK_GROUP_SIZE";                   break; }
		case CL_INVALID_WORK_ITEM_SIZE:                    { message = "CL_INVALID_WORK_ITEM_SIZE";                    break; }
		case CL_INVALID_GLOBAL_OFFSET:                     { message = "CL_INVALID_GLOBAL_OFFSET";                     break; }
		case CL_INVALID_EVENT_WAIT_LIST:                   { message = "CL_INVALID_EVENT_WAIT_LIST";                   break; }
		case CL_INVALID_EVENT:                             { message = "CL_INVALID_EVENT";                             break; }
		case CL_INVALID_OPERATION:                         { message = "CL_INVALID_OPERATION";                         break; }
		case CL_INVALID_GL_OBJECT:                         { message = "CL_INVALID_GL_OBJECT";                         break; }
		case CL_INVALID_BUFFER_SIZE:                       { message = "CL_INVALID_BUFFER_SIZE";                       break; }
		case CL_INVALID_MIP_LEVEL:                         { message = "CL_INVALID_MIP_LEVEL";                         break; }
		case CL_INVALID_GLOBAL_WORK_SIZE:                  { message = "CL_INVALID_GLOBAL_WORK_SIZE";                  break; }
		case CL_INVALID_PROPERTY:                          { message = "CL_INVALID_PROPERTY";                          break; }
	}

	std::stringstream stream;
	stream << file << " - line " << line << std::endl << '\t' << CollapseWhitespace( what ) << std::endl << '\t' << message << " (code=" << code << ')';
	m_what = stream.str();
}


Exception::~Exception() throw() {
}


const char* Exception::what() const throw() {

	return m_what.c_str();
}




}    // namespace OpenCL




#endif    // SETTING_OPENCL
