#pragma once

#ifndef OPENCL_PROGRAM_H
#define OPENCL_PROGRAM_H


/**
	\file
	\brief Contains OpenCL::Program class.
*/




#include "settings.h"




#if ( SETTING_OPENCL != 0 )




#include "opencl_state.h"
#include "opencl_deleter.h"

#include <CL/cl.h>

#include <memory>




namespace OpenCL {




/*==============================================================================
	Program class
==============================================================================*/


/**
	\brief OpenCL program class.

	The constructor attempts to create and build a program from a list of
	source files. If any of the OpenCL calls which perform this task fail, than
	an OpenCL Exception will be thrown. The compiled program can be accessed
	via the \c BuiltProgram accessor function.
*/
struct Program {

	/**
		\brief Constructs and builds an OpenCL program.

		\param pState       Pointer to the OpenCL State for which to build this program.
		\param filenames    <tt>NULL</tt>-terminated list of <tt>'\0'</tt>-terminated strings containing the files to build (each containing OpenCL source).
		\param options      Compiler options to pass to \c clBuildProgram.
		\param logFilename  File to which to write a build log, on failure. If NULL, no log will be written.
	*/
	Program(
		std::shared_ptr< State > const& pState,
		char const* const* const filenames,
		char const* const options,
		char const* const logFilename = NULL
	);


	inline cl_program BuiltProgram() const;    ///< \brief Returns built OpenCL program.


private:

	std::shared_ptr< State > m_pState;
	std::unique_ptr< _cl_program, Deleter< _cl_program > > m_program;
};




/*==============================================================================
	Program methods
==============================================================================*/


cl_program Program::BuiltProgram() const {

	return m_program.get();
}




}    // namespace OpenCL




#endif    // SETTING_OPENCL




#endif    // OPENCL_PROGRAM_H
